#ifndef COM_SAXBOPHONE_CIRCULAR_QUEUE_BUFFER_HPP
#define COM_SAXBOPHONE_CIRCULAR_QUEUE_BUFFER_HPP

/**
 * @file
 *
 * @brief A templated class type implementing a simple circular queue buffer
 *
 * @details This is useful in many places in the project to implement simple
 * buffers, the contents of which don't need shifting as items are removed from
 * the head. This type automatically handles index-translation.
 * @note As this is a templated type, this is a header-only compilation unit.
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include <cstddef> // size_t


// SAXBOPHONE.COM is my domain, so I'm using the reverse form as my namespace
namespace com_saxbophone {
    template<
        typename T, /**< the type being stored */
        size_t BUFFER_SIZE = 16u /**< default buffer size */
    >
    class CircularQueueBuffer {
    public:
        /**
         * @brief Default Constructor
         */
        CircularQueueBuffer()
            : buffer_items() // zero-initialise all members
            , buffer_head(0)
            , buffer_length(0)
            {}

        /**
         * @brief Tries to add a new item to the end of the buffer
         * @returns true if item was added successfully
         * @returns false if the buffer is full and item could not be added
         */
        bool add(T item) {
            // guard clause to prevent adding to a full buffer
            if (this->is_full()) return false;
            // if it's not full, then store the item
            this->buffer_items[this->next_index()] = item;
            // and update the buffer book-keeping
            this->buffer_length++;
            return true; // done successfully
        }

        /**
         * @brief Overloaded [] operator to allow indexing into the buffer in
         * an array-like fashion, for read-only purposes only
         * @returns the item at the given index
         * @warn Don't try and retrieve items with invalid indexes!
         */
        T operator[](size_t i) const {
            return this->buffer_items[this->translate_index(i)];
        }

        /**
         * @returns The current length of this buffer
         * @note In this context, "length" refers to how many items are
         * currently stored. To retrieve the total capacity of the buffer,
         * see `capacity()`.
         */
        size_t length() const {
            return this->buffer_length;
        }

        /**
         * @returns the total capacity of this buffer
         */
        constexpr size_t capacity() const {
            return BUFFER_SIZE;
        }

        /**
         * @returns whether this buffer is empty or not
         */
        bool is_empty() const {
            return this->buffer_length == 0;
        }

        /**
         * @returns whether this buffer is full or not
         */
        bool is_full() const {
            return this->buffer_length == BUFFER_SIZE;
        }

        /**
         * @brief Empties the entire contents of the buffer
         * @note If you know you want to empty the buffer, this is more
         * efficient than calling shrink_from_front()
         */
        void clear() {
            this->buffer_head = 0;
            this->buffer_length = 0;
        }

        /**
         * @brief Shrinks the buffer by removing up to `count` number of items
         * from the front of it, moving the head of the buffer to the position
         * after the last removed item and setting buffer length accordingly
         * @returns true if buffer was successfully shrunk
         * @returns false if buffer could not be shrunk by the given number of
         * elements, because it has fewer elements than the requested amount
         */
        bool shrink_from_front(size_t count) {
            // check count is not greater than buffer size before continuing
            if (count > this->buffer_length) {
                return false;
            }
            // optimisation: if full shrink is requested, just clear it
            if (count == this->buffer_length) {
                this->clear();
                return true; // this can't fail
            }
            /*
             * otherwise, continue and shrink the buffer
             * NOTE: shrinking just requires changing head and length
             * --no objects are destroyed
             */
            // the head of the buffer is moved forwards, respecting wrap-around
            this->buffer_head = (this->buffer_head + count) % BUFFER_SIZE;
            // the length is simply reduced by the number of elements
            this->buffer_length -= count;
            return true; // all done
        }
    private:
        /**
         * @returns the index of the first empty location within the transition
         * buffer
         */
        size_t next_index() const {
            // item at index buffer_length is the first item outside the buffer
            return this->translate_index(this->buffer_length);
        }

        /**
         * @returns the raw index of the given 0-index into the circular buffer
         * @details Because the buffer is circular and isn't shifted around much,
         * item at index 0 in the buffer as we visualise it might be at a
         * different location than index 0 in memory. This method provides this
         * translation.
         * @warn Don't use this method with indexes at or above buffer_length
         * unless you know what you're doing
         * @warn Never use this method with indexes at or above BUFFER_SIZE
         */
        size_t translate_index(size_t i) const {
            /*
             * since this is a circular buffer, the tail of it will wrap around
             * the end of the array in which it is stored eventually, so we need
             * to account for that with some modulo arithmetic
             */
            return (this->buffer_head + i) % BUFFER_SIZE;
        }

        /**
         * @brief Circular-buffer style queue used for storing items of type T.
         * @details The size of this can be configured by changing the optional
         * template non-type paramter BUFFER_SIZE to a different value. The size
         * ought to be kept reasonably small to prevent excessive RAM usage on
         * the RAM-constrained micro:bit, however a buffer size that is too
         * small may cause premature capacity exhaustion in certain use cases.
         */
        T buffer_items[BUFFER_SIZE];
        /**
         * @brief the current index of the first item in buffer
         */
        size_t buffer_head;
        /**
         * @brief the current number of items stored in the buffer
         */
        size_t buffer_length;
    };
};

#endif // include guard
