#ifndef COM_SAXBOPHONE_MANCHESTER_DECODER_HPP
#define COM_SAXBOPHONE_MANCHESTER_DECODER_HPP

/**
 * @file
 *
 * @brief A simple class that performs Manchester Decoding of input line states.
 *
 * @details HI/LOW states are sent into the decoder with the input() method and
 * decoded bytes are returned with the get_next_byte() method.
 * Bytes are decoded in big-endian order, from a Manchester-encoded signal
 * that conforms to that used in IEEE 802.3, that is, the clock cycles HI-LOW
 * and output is Clock XOR Data. Thus, a HI-LOW transition indicates binary zero,
 * whereas a LOW-HI transition indicates binary one.
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include <cstdint> // uint8_t

#include "CircularQueueBuffer.hpp"


// SAXBOPHONE.COM is my domain, so I'm using the reverse form as my namespace
namespace com_saxbophone {
    class ManchesterDecoder {
    public:
        /**
         * @brief Default constructor
         */
        ManchesterDecoder();

        /**
         * @brief Read in another state from the line to be decoded.
         * @returns true if the state has been accepted
         * @returns false if the state has not been accepted (buffer full)
         */
        bool input(bool state);

        /**
         * @brief Decodes the next byte from input (if available) and if there
         * is a byte available, stores it in `destination`.
         * @param destination[out] the byte to write the next decoded byte to
         * @returns true if there was a byte available and it was stored
         * @returns false if no bytes are available yet (destination is ignored)
         */
        bool get_next_byte(uint8_t& destination);

        /**
         * @brief Clears the input buffer, in case you want to do that.
         */
        void clear();
    private:
        /**
         * @brief Counter tracking how many "symbols" (i.e. boolean input states)
         * we have seen so far
         * @details This wraps around at 16. 8 bits is represented as 16 symbols
         * because of the integrated clock signal --as this implementation is
         * IEEE 802.3, every second symbol holds the value of a decoded data bit
         */
        uint8_t symbol_index;

        /**
         * @brief Internal circular-buffer to cache decoded output bytes
         * @detail Once the current_byte is full, it is pushed into this buffer,
         * ready to be retrieved with get_next_byte()
         */
        CircularQueueBuffer<uint8_t> buffer;

        /**
         * @brief The current byte we are populating from decoded input bits
         * @details This is used as a temporary cache to store the partially
         * constructed input byte. Once the byte is full (all bits stored), it
         * is added to the end of the buffer.
         */
        uint8_t current_byte;
    };
};

#endif // include guard
