/**
 * @file
 *
 * @brief Implementation of class ManchesterEncoder
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include <cstdint> // uint8_t

#include "ManchesterEncoder.hpp"


// SAXBOPHONE.COM is my domain, so I'm using the reverse form as my namespace
namespace com_saxbophone {
    ManchesterEncoder::ManchesterEncoder()
        : clock(true) // our implementation has the clock signal starting HI
        , bit_index(0) // we definitely want to start reading bit 0
        {}

    bool ManchesterEncoder::input(uint8_t byte) {
        return this->buffer.add(byte); // buffer will return false if it's full
    }

    bool ManchesterEncoder::get_next_state(int& destination) {
        // guard condition --check if buffer has content
        if (this->buffer.is_empty()) return false;
        // Manchester Code = Clock XOR Data
        int symbol = this->clock xor this->extract_bit(
            this->buffer[0],
            this->bit_index
        );
        destination = symbol;
        // if we are at the end of a byte
        if (this->clock == false and this->bit_index == 7) {
            // shrink the buffer from the front by 1 to advance the buffer
            this->buffer.shrink_from_front(1);
        }
        // update bit index whenever clock is LOW (end of duration of 1 data bit)
        if (not this->clock) this->bit_index = (this->bit_index + 1u) % 8u;
        // update clock every time
        this->clock = not this->clock;
        return true;
    }

    bool ManchesterEncoder::extract_bit(uint8_t byte, uint8_t bit_index) const {
        // NOTE: this is a big-endian extraction
        uint8_t mask = 0x80u >> bit_index;
        return (byte & mask); // no need to shift down, anything non-zero is true
    }
};
