#ifndef COM_SAXBOPHONE_MANCHESTER_ENCODER_HPP
#define COM_SAXBOPHONE_MANCHESTER_ENCODER_HPP

/**
 * @file
 *
 * @brief A simple class that performs Manchester Encoding of input bytes
 *
 * @details Bytes are sent into the encoder with the input() method and line
 * states to set (HI/LO) are returned with the get_next_state() method.
 * Bytes are encoded in big-endian order and the Manchester Encoding version
 * conforms to that used in IEEE 802.3, that is, the clock cycles HI-LOW and
 * output is Clock XOR Data. Thus, a HI-LOW transition indicates binary zero,
 * whereas a LOW-HI transition indicates binary one.
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include <cstdint> // uint8_t

#include "CircularQueueBuffer.hpp"


// SAXBOPHONE.COM is my domain, so I'm using the reverse form as my namespace
namespace com_saxbophone {
    class ManchesterEncoder {
    public:
        /**
         * @brief Default constructor
         */
        ManchesterEncoder();

        /**
         * @brief Read in another byte of input to be encoded by the encoder
         * @returns true if the byte has been accepted
         * @returns false if the byte has not been accepted (buffer full)
         */
        bool input(uint8_t byte);

        /**
         * @brief retrieves the next state that the encoder wants the output
         * line to be set to, where 0 = LOW and 1 = HIGH, stores it in
         * `destination`.
         * @param destination[out] the int to write the next state to
         * @returns true if there was a state available and it was stored
         * @returns false if no states are available yet (destination is ignored)
         */
        bool get_next_state(int& destination);
    private:
        /**
         * @brief Extracts the bit at given index from the given byte
         * @note this uses a big-endian bit addressing system
         */
        bool extract_bit(uint8_t byte, uint8_t bit_index) const;

        /**
         * @brief Clock flag used to track this ManchesterEncoder's clock signal
         * @details The clock signal of a Manchester Encoder alternates at twice
         * the rate of data. In our implementation, our clock alternates HI-LO
         * in that order.
         */
        bool clock;

        /**
         * @brief Which bit of the byte of data we are currently on
         * @details This cycles in the range {0..7} and tells us which bit to
         * read out of the byte at the head of the buffer next. When it
         * overflows, it's time to advance the buffer and read the next byte
         */
        uint8_t bit_index;

        /**
         * @brief Internal circular-buffer to cache input bytes
         */
        CircularQueueBuffer<uint8_t> buffer;

        /**
         * @brief The current byte we have started sending
         * @details Bytes are extracted from the buffer into here. If the buffer
         * is ever exhausted, 0x00 is placed here to give the encoder something
         * to encode.
         */
        uint8_t current_byte;
    };
};

#endif // include guard
