#ifndef COM_SAXBOPHONE_MORSE_DECODER_HPP
#define COM_SAXBOPHONE_MORSE_DECODER_HPP

/**
 * @file
 *
 * @brief A simple class that can be used to turn sequences of times which a
 * button was held/released into Morse Code output.
 *
 * @details TODO ADD DESCRIPTION HERE
 *
 * @todo Move the implementation code out of this header and into a
 * corresponding .cpp file. I can't do this right now as I currently don't know
 * how to make Yotta build multiple objects...
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include <cstddef> // size_t

#include "CircularQueueBuffer.hpp"
#include "MorseSymbol.hpp"
#include "MorseTree.hpp"
#include "Pulse.hpp"


// SAXBOPHONE.COM is my domain, so I'm using the reverse form as my namespace
namespace com_saxbophone {
    class MorseDecoder {
    public:
        /**
         * @brief Constructor
         * @param unit_duration the duration of one Morse Code 'unit' in ms
         * @todo Consider converting to constexpr to allow const derivation of
         * unit durations at compile-time.
         */
        MorseDecoder(unsigned long unit_duration)
            // pre-compute the duration of symbols in ms
            : dot_duration(unit_duration * MorseDecoder::DOT_DURATION)
            , dash_duration(unit_duration * MorseDecoder::DASH_DURATION)
            , space_duration(unit_duration * MorseDecoder::SPACE_DURATION)
            , letter_space_duration(unit_duration * MorseDecoder::LETTER_SPACE_DURATION)
            , word_space_duration(unit_duration * MorseDecoder::WORD_SPACE_DURATION)
            , unit_duration(unit_duration)
            {}

        /**
         * @brief Call this method after the input control has been held or
         * released, giving the duration for which it was held or released in ms
         * @param held whether the input control was held down or not
         * @param duration the duration in ms for which the input was held or
         * released
         * @returns false if value of held is the same as that of the last call
         * to this method (you can't have held down the button on two separate
         * occasions with no pause in-between!). This can also happen if the
         * buffer overflows. In either case, the MorseDecoder instance will no
         * longer function and should be re-instantiated.
         * @todo consider changing return type to an enum to allow detecting
         * these two error conditions and the success condition.
         * @returns true on receipt of a valid input transition
         * @note if input is noise (i.e. too short), this method still returns
         * true but input is ignored
         */
        bool transition(bool held, unsigned long duration) {
            // before we accept this input, we need to check we can use it...
            if (duration < this->unit_duration) { // is it noise?
               return true; // do nothing, but it's not an error condition
            }
            // if the held state is the same as the last seen, this is an error
            if (held == this->last_seen) {
                return false;
            }
            /*
             * TODO: do we need to check if it's the first input and low?
             * --probably not as the guard-if above should stop that too, as
             * this->last_seen will initially be false and we will set it to
             * false every time end of input is detected.
             */
            // if execution reaches this point, we know we need to store
            // try and store a new Pulse constructed from these params
            Pulse p = { held, duration, };
            if (not this->buffer.add(p)) {
                // oh no, a buffer *would* have overflown, but we won't let it!
                return false;
            }
            // if execution reached here, we know store was successful
            this->last_seen = held; // update last seen
            // TODO: do we have to do anything else at this point? Probably not
            return true;
        }

        /**
         * @brief Attempts to remove the first available decoded character from
         * the internal queue of decoded characters maintained by the
         * MorseDecoder.
         * @details If a decoded character is available, this is removed from
         * the head of the internal queue and stored in the referenced
         * destination
         * @param[out] destination the char variable to store any removed
         * character in
         * @returns true if a decoded character was removed from the internal
         * queue and stored
         * @returns false if no fully decoded characters are yet available
         */
        bool remove_next(char& destination) {
            /*
             * TODO: implement adaptive decoding
             * NOTE: to make decoding easier, you can take the average length of
             * the shorter spaces to get some kind of metric for how long 1 unit
             * is (and reject long spaces by checking if they're around 300% of
             * our current average value for one space). Care must be taken to
             * make sure we can tell apart the first two spaces (these should be
             * compared directly to see if one is ~300% of the other)
             */
            /*
             * the current decoder implementation just uses threshold timing to
             * determine dots from dashes and the like
             */
            /*
             * the minimum number of transitions we need for a symbol is 2
             * (E and T consist of one pulse, so that'll be one on and one off)
             */
            if (this->buffer.length() < 2) {
                return false; // not enough content in the buffer yet to decode
            }
            // use the MorseTree to try and decode some morse
            /*
             * we also want to know how many pulses were read before a letter
             * boundary was encountered, so pulses_read will track this, and if
             * the method found a suitable ending Pulse, so we know if this
             * input sequence is ready to be fully decoded
             */
            size_t pulses_read = 0;
            bool found_end = false;
            MorseTree decoder_tree = this->read_buffer_into_tree(
                &pulses_read,
                &found_end
            );
            // if a suitable ending Pulse was not found, there's no output
            if (not found_end) {
                // NOTE: we don't shrink the buffer here, we'll need it later
                return false;
            }
            // in all cases, we now need to remove the parsed pulses from buffer
            this->buffer.shrink_from_front(pulses_read);
            // now read the tree's cursor, then we can parse it
            char symbol = decoder_tree.read_cursor();
            switch (symbol) {
            case '!':
                // NOTE: **DELIBERATE FALLTHROUGH**
            case '?':
                /*
                 * - if exclamation mark is returned, the input was invalid
                 * - if question mark is returned, input is not complete
                 * --if either are the case, then there is no output to return
                 */
                return false;
            default:
                // in any other case, we have a valid symbol --store in out param
                destination = symbol;
                return true;
            }
        }

        /**
         * @returns whatever letter is represented by the sequence entered so
         * far
         * @returns '?' when the sequence entered so far is not a valid Morse
         * code on its own, but could be if more is added to the sequence
         * @returns '!' when the sequence entered is not a valid Morse code and
         * does not prefix any other valid Morse code.
         * @note This method does not alter the state of the internal queue
         */
        char peek() const {
            // send what's in the buffer into a tree and return cursor value
            return this->read_buffer_into_tree(nullptr, nullptr).read_cursor();
        }

        /**
         * @returns a suggested duration to wait before considering input to
         * have finished for the time being, measured in ms.
         * @details this is provided to prevent the program from hanging on
         * input conclusion, as the decoding system is transition-based and
         * would be eternally waiting for the next input transition otherwise
         * once input had finished, if such timeout behaviour was not provided.
         * @note this method will use either the value of max_duration, or a
         * different value based on the decoder's understanding of how fast the
         * user 'types' Morse Code.
         */
        unsigned long suggested_timeout_duration() const {
            // this decoder is not currently adaptive, so return a fixed value
            return this->word_space_duration;
        };

    private:
        /**
         * @brief Attempts to read transitions from the buffer, converting them
         * to Pulses and sending them into a MorseTree, which is returned
         * @param[out] pulses_read optional pointer to variable which will count
         * how many Pulses this method read into the tree
         * @param[out] found_end optional flag specifying whether this method
         * found a LETTER SPACE or WORD SPACE pulse (it will stop decoding there
         * if it did)
         */
        MorseTree read_buffer_into_tree(
            size_t* pulses_read,
            bool* found_end
        ) const {
            bool stop = false;
            // instantiate a MorseTree decoder we will use to try and decode
            MorseTree decoder_tree;
            // keep reading until we either exhaust the buffer or reach a space
            size_t i; // we'll re-use this after looping to count how pulses
            for (i = 0; not stop and i < this->buffer.length(); i++) {
                // retrieve the Pulse from buffer
                Pulse input = this->buffer[i];
                // convert it to a Morse symbol
                MorseSymbol symbol = this->convert_transition_to_morse(input);
                // decide what to do with the symbol
                switch (symbol) {
                    // Dots and Dashes are sent to the tree
                case MorseSymbol::DOT:
                    decoder_tree.advance(false);
                    break;
                case MorseSymbol::DASH:
                    decoder_tree.advance(true);
                    break;
                    // LETTER SPACE or WORD SPACE means we can finish
                case MorseSymbol::LETTER_SPACE:
                    // **DELIBERATE FALLTRHOUGH**
                case MorseSymbol::WORD_SPACE:
                    stop = true; // we can now stop
                    break;
                }
            }
            // set "found_end" if it's been given to us
            if (found_end != nullptr) {
                *found_end = stop;
            }
            // set "pulses_read" if it's been given to us
            if (pulses_read != nullptr) {
                *pulses_read = i;
            }
            return decoder_tree;
        }

        /**
         * @returns A MorseSymbol decoded from the given Pulse.
         */
        MorseSymbol convert_transition_to_morse(Pulse transition) const {
            // first of all, distinguish between counting marks or spaces
            if (transition.state) { // it's a dot or a dash ("mark")
                /*
                 * across the alphanumeric range of International Morse, Dots
                 * are slightly more common than Dashes (68 Dots vs 63 Dashes
                 * seen across the entire set), so we'll prioritise Dots
                 */
                if (transition.duration < this->dash_duration) { // could be Dot?
                    if (transition.duration >= this->dot_duration) { // is it Dot?
                        return MorseSymbol::DOT;
                    }
                    // otherwise, continue to default "Unknown" condition at end
                } else { // otherwise, it's a Dash
                    return MorseSymbol::DASH;
                }
            } else { // it's a space --but what kind?
                // inter-character spaces are the most common, so prioritise them
                if (transition.duration < this->letter_space_duration) {
                    // it could be an inter-character space
                    if (transition.duration >= this->space_duration) {
                        return MorseSymbol::SPACE;
                    }
                    // otherwise, continue to default "Unknown" condition at end
                } else { // it's either a space between letters or words
                    // is it a space between letters?
                    if (transition.duration < this->word_space_duration) {
                        return MorseSymbol::LETTER_SPACE; // it's a Letter Space
                    } else {
                        return MorseSymbol::WORD_SPACE; // it's a Word Space
                    }
                }
            }
            // if execution reached here, we don't known what it is
            return MorseSymbol::UNKNOWN;
        }

        /**
         * @brief The number of units long that a Morse Code "DOT" is
         */
        const static unsigned long DOT_DURATION = 1u;
        /**
         * @brief The number of units long that a Morse Code "DASH" is
         */
        const static unsigned long DASH_DURATION = 3u;
        /**
         * @brief The number of units long that the space between parts of the
         * same letter is
         */
        const static unsigned long SPACE_DURATION = 1u;
        /**
         * @brief The number of units long that the space between letters is
         */
        const static unsigned long LETTER_SPACE_DURATION = 3u;
        /**
         * @brief The number of units long that the space between words is
         */
        const static unsigned long WORD_SPACE_DURATION = 7u;

        /**
         * @brief Pre-computed value of the duration of a Dot in ms
         */
        unsigned long dot_duration;

        /**
         * @brief Pre-computed value of the duration of a Dash in ms
         */
        unsigned long dash_duration;

        /**
         * @brief Pre-computed value of the duration of a Space in ms
         */
        unsigned long space_duration;

        /**
         * @brief Pre-computed value of the duration of a Letter Space in ms
         */
        unsigned long letter_space_duration;

        /**
         * @brief Pre-computed value of the duration of a Words Space in ms
         */
        unsigned long word_space_duration;

        /**
         * @brief Circular-buffer style queue used to store transitions as
         * passed into this decoder when transition() is called.
         * @details The size of this can be configured by changing the optional
         * second template non-type paramter to a different value. The size
         * ought to be kept reasonably small to prevent excessive RAM usage on
         * the RAM-constrained micro:bit, however a buffer size that is too
         * small may leave the MorseDecoder unable to function!
         */
        CircularQueueBuffer<Pulse> buffer;

        /**
         * @brief the duration of one Morse Code "unit" in ms
         */
        unsigned long unit_duration;

        /**
         * @brief stores true if last input seen was 'held', false otherwise
         */
        bool last_seen;
    };
};

#endif // include guard
