#ifndef COM_SAXBOPHONE_MORSE_TREE_HPP
#define COM_SAXBOPHONE_MORSE_TREE_HPP

/**
 * @file
 *
 * @brief A simple struct type used to decode Morse Code via a tree structure
 *
 * @details TODO ADD DESCRIPTION HERE
 *
 * @todo Move the implementation code out of this header and into a
 * corresponding .cpp file. I can't do this right now as I currently don't know
 * how to make Yotta build multiple objects...
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "MorseSymbol.hpp"


// SAXBOPHONE.COM is my domain, so I'm using the reverse form as my namespace
namespace com_saxbophone {
    /**
     * @brief One leaf/node of the Morse-decoding tree
     * @note On a RAM-constrained platform such as the micro:bit, this type is
     * designed to be stored const so that it will be placed in the FLASH memory
     */
    struct MorseTreeNode {
        /**
         * @brief The character that this node represents, if any
         * @note '?' represents an indeterminate character value at that node
         */
        char symbol;
        /**
         * @brief Pointer to the next node to traverse if the next Morse Symbol
         * is a Dot (can be NULL)
         */
        const MorseTreeNode* dot;
        /**
         * @brief Pointer to the next node to traverse if the next Morse Symbol
         * is a Dash (can be NULL)
         */
        const MorseTreeNode* dash;
    };

    /*
     * it is intended that these constants be stored in FLASH memory by the
     * compiler/linker in the interest of conserving the limited RAM that
     * the micro:bit has available. Because there are pointers between many
     * of these variables, they have to be declared and initialised in a
     * very specific order.
     *
     * NOTE: nodes named `STEM_<n>` are not themselves legal letters, but
     * provide routes towards them.
     */
    // these are all the leaf nodes
    /* ********************************** char DOT?     DASH?    */ // notation
    static const MorseTreeNode NODE_5 = { '5', nullptr, nullptr, }; // .....
    static const MorseTreeNode NODE_4 = { '4', nullptr, nullptr, }; // ...._
    static const MorseTreeNode NODE_3 = { '3', nullptr, nullptr, }; // ...__
    static const MorseTreeNode NODE_2 = { '2', nullptr, nullptr, }; // ..___
    static const MorseTreeNode NODE_1 = { '1', nullptr, nullptr, }; // .____
    static const MorseTreeNode NODE_6 = { '6', nullptr, nullptr, }; // _....
    static const MorseTreeNode NODE_7 = { '7', nullptr, nullptr, }; // __...
    static const MorseTreeNode NODE_8 = { '8', nullptr, nullptr, }; // ___..
    static const MorseTreeNode NODE_9 = { '9', nullptr, nullptr, }; // ____.
    static const MorseTreeNode NODE_0 = { '0', nullptr, nullptr, }; // _____
    // next come the intermediate nodes
    static const MorseTreeNode NODE_H = { 'H', &NODE_5, &NODE_4, }; // ....
    static const MorseTreeNode NODE_V = { 'V', nullptr, &NODE_3, }; // ..._
    static const MorseTreeNode NODE_F = { 'F', nullptr, nullptr, }; // .._.
    static const MorseTreeNode STEM_1 = { '?', nullptr, &NODE_2, }; // ..__
    static const MorseTreeNode NODE_L = { 'L', nullptr, nullptr, }; // ._..
    static const MorseTreeNode NODE_P = { 'P', nullptr, nullptr, }; // .__.
    static const MorseTreeNode NODE_J = { 'J', nullptr, &NODE_1, }; // .___
    static const MorseTreeNode NODE_B = { 'B', &NODE_6, nullptr, }; // _...
    static const MorseTreeNode NODE_X = { 'X', nullptr, nullptr, }; // _.._
    static const MorseTreeNode NODE_C = { 'C', nullptr, nullptr, }; // _._.
    static const MorseTreeNode NODE_Y = { 'Y', nullptr, nullptr, }; // _.__
    static const MorseTreeNode NODE_Z = { 'Z', &NODE_7, nullptr, }; // __..
    static const MorseTreeNode NODE_Q = { 'Q', nullptr, nullptr, }; // __._
    static const MorseTreeNode STEM_0 = { '?', &NODE_8, nullptr, }; // ___.
    static const MorseTreeNode STEM_2 = { '?', &NODE_9, &NODE_0, }; // ____
    // one level up
    static const MorseTreeNode NODE_S = { 'S', &NODE_H, &NODE_V, }; // ...
    static const MorseTreeNode NODE_U = { 'U', &NODE_F, &STEM_1, }; // .._
    static const MorseTreeNode NODE_R = { 'R', &NODE_L, nullptr, }; // ._.
    static const MorseTreeNode NODE_W = { 'W', &NODE_P, &NODE_J, }; // .__
    static const MorseTreeNode NODE_D = { 'D', &NODE_B, &NODE_X, }; // _..
    static const MorseTreeNode NODE_K = { 'K', &NODE_C, &NODE_Y, }; // _._
    static const MorseTreeNode NODE_G = { 'G', &NODE_Z, &NODE_Q, }; // __.
    static const MorseTreeNode NODE_O = { 'O', &STEM_0, &STEM_2, }; // ___
    // one level up
    static const MorseTreeNode NODE_I = { 'I', &NODE_S, &NODE_U, }; // ..
    static const MorseTreeNode NODE_A = { 'A', &NODE_R, &NODE_W, }; // ._
    static const MorseTreeNode NODE_N = { 'N', &NODE_D, &NODE_K, }; // _.
    static const MorseTreeNode NODE_M = { 'M', &NODE_G, &NODE_O, }; // __
    // one level up
    static const MorseTreeNode NODE_E = { 'E', &NODE_I, &NODE_A, }; // .
    static const MorseTreeNode NODE_T = { 'T', &NODE_N, &NODE_M, }; // _
    // here is the root node
    static const MorseTreeNode ROOT   = { '?', &NODE_E, &NODE_T, }; //

    class MorseTree {
    public:
        /**
         * @brief Default constructor
         */
        MorseTree()
            : cursor(&ROOT) // cursor always starts at the root
            , am_lost(false)
            {}

        /**
         * @brief Sends another Dot or Dash move instruction to the Tree's cursor
         * @param is_dash Set this to false if you want to send a Dot, or true
         * if you want to send a Dash
         */
        void advance(bool is_dash) {
            // refuse to do anything if we're lost
            if (not this->am_lost) {
                // otherwise, let's traverse the tree
                this->cursor = is_dash ? this->cursor->dash : this->cursor->dot;
                // detect null pointers
                if (this->cursor == nullptr) {
                    // we consider ourselves "lost" if we walk off of the tree
                    this->am_lost = true;
                }
            }
        }

        /**
         * @returns The current character that the cursor is looking at in the
         * Tree
         * @returns '!' if the sequence of Dots and Dashes that has previously
         * been sent to the tree is not a valid Morse code sequence nor a prefix
         * of one
         * @returns '?' if the sequence of Dots and Dashes that has previously
         * been sent to the tree is not valid on its own but does prefix one or
         * more valid Morse codes.
         */
        char read_cursor() const {
            // if we're lost, return '!', otherwise use current cursor symbol
            return this->am_lost ? '!' : this->cursor->symbol;
        }

    private:
        /*
         * this is our cursor, it tracks where we are currently in the tree and
         * always starts at the root
         */
        const MorseTreeNode* cursor;
        // this flag tracks if the cursor "got lost", by travelling off the tree
        bool am_lost;
    };
}

#endif // include guard
