/**
 * @file
 *
 * @brief This is my submission for Coursework Challenge 2 of UWE's Internet of
 * Things module.
 *
 * @details TODO ADD DESCRIPTION HERE
 *
 * @author Joshua Saxby <Joshua2.Saxby@live.uwe.ac.uk>
 * @date 2020
 *
 * Student Number 18018052
 *
 * @copyright Copyright (C) Joshua Saxby 2020
 *
 * @copyright
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "MicroBit.h"

#include "ManchesterDecoder.hpp"
#include "ManchesterEncoder.hpp"
#include "MorseDecoder.hpp"
#include "Pulse.hpp"
#include "Role.hpp"
#include "Transceiver.hpp"
#include "TransceiverConfig.hpp"


// special config macros
#define MORSE_UNIT_MS 100 // One Morse "unit" is 100ms duration
// micro:bit pin number and pin name --important: both of these must correspond!
#define COMMS_PIN_NUMBER 1
#define COMMS_PIN_NAME MICROBIT_PIN_P1
#define COMMS_PIN_EVENT_ID MICROBIT_ID_IO_P1


// please import all names from my namespace into local scope!
using namespace com_saxbophone;

// our single global MicroBit instance
MicroBit ubit;


/**
 * @brief Simple function for just working out whether this micro:bit is going
 * to be the Master or Slave for communications purposes
 * @warn This method blocks until exactly one of button A or B is pressed!
 */
Role get_my_role() {
    // ubit.serial.send("get_my_role()");
    // scroll some stuff on the screen to prompt the user to choose
    ubit.display.scroll("<M"); // point left to Button A to choose Master
    ubit.sleep(100); // a brief pause
    ubit.display.scroll("S>"); // point right to Button B to choose Slave
    bool a_pressed = false;
    bool b_pressed = false;
    do {
        ubit.sleep(10); // give the CPU a little bit of a break
        a_pressed = ubit.buttonA.isPressed();
        b_pressed = ubit.buttonB.isPressed();
    // NOTE: this is bitwise-XOR but given the boolean nature, it's acceptable
    } while (not (a_pressed xor b_pressed)); // until exactly one button pressed
    // because of the XOR exit condition of loop, if a_pressed = 0, b_pressed = 1
    return a_pressed ? Role::MASTER : Role::SLAVE; // A = MASTER, B = SLAVE
}


class Pairer {
public:
    /**
     * @brief Default constructor
     */
    Pairer()
        : synchronisation_timestamp(0) // zero-initialise to prevent any issues
        {}

    /**
     * @brief Attempt to pair with another connected micro:bit
     * @param my_role Whether we wish to be the Master or Slave for comms
     * purposes
     * @returns 0 if paired successfully
     * @returns error code number if unable to pair successfully
     * @note This is a blocking call!
     */
    int attempt_pair(Role my_role) {
        if (my_role == Role::MASTER) {
            // attempt pair as Master
            return this->pair_as_master();
        } else if (my_role == Role::SLAVE) {
            // attempt pair as Slave
            return this->pair_as_slave();
        } else {
            // invalid Role!
            return -1;
        }
    }

    /**
     * @brief Returns to the caller once the amount of time since synchronisation
     * reaches a multiple of the transmission window.
     * @details This is intended to be used to keep both Master and Slave
     * synchronised. Assuming their clocks were synchronised correctly, both can
     * call this and the calls should return at near enough the same point in time.
     */
    void barrier() {
        // TODO: consider allowing for a ±10% margin of error
        unsigned long start_time = ubit.systemTime(); // what time is it now?
        // this is how much time we have left to next sync point
        unsigned long time_left = TransceiverConfig::TRANSMISSION_WINDOW - (
            (start_time - this->synchronisation_timestamp) %
            TransceiverConfig::TRANSMISSION_WINDOW
        );
        // calculate the time of the next sync point
        unsigned long sync_point = start_time + time_left;
        /*
         * systemTime() can only sleep to 6ms resolution, so sleep to the nearest
         * multiple of 6ms, rounded down
         */
        ubit.sleep((time_left / 6) * 6);
        // busy wait loop sleeps for remainder of time
        unsigned long time_now;
        do {
            time_now = ubit.systemTime();
        } while (time_now < sync_point);
    }

private:
    int pair_as_master() {
        int pin_state = ubit.io.pin[COMMS_PIN_NUMBER].getDigitalValue();
        // clarify if someone else is driving the pin HI or if pin is invalid
        switch (pin_state) {
        case 0:
            break; // this is good, it means no-one else is driving the pin
        case 1:
        /* DELIBERATE FALLTHROUGH */
        case MICROBIT_NOT_SUPPORTED:
            /*
             * either someone else on the line thinks they're Master, or we've
             * been configured with a pin that can't do Digital I/O. In either
             * case, this is a failure condition.
             */
            // ubit.serial.send("MICROBIT_NOT_SUPPORTED\r\n");
            return 2;
        }
        // successful execution continues here
        // wait for the pin to stay LO for 500ms
        unsigned long start_time, end_time;
        this->busy_wait_on_pin(0, 500, pin_state, start_time, end_time);
        // check to make sure we didn't break out of the loop because of HI pin
        if (pin_state != 0) {
            // ubit.serial.send("Other Master on Line\r\n");
            return 4;
        }
        // there's definitely no-one else transmitting on the line, now we transmit
        this->busy_wait_drive_pin(500, start_time, end_time);
        // end_time is the clock synchronisation timestamp
        this->synchronisation_timestamp = end_time;
        // now, wait up to 200ms for a reply from the Slave
        this->busy_wait_on_pin(0, 200, pin_state, start_time, end_time);
        // this time, check that the pin is HI --it must be HI, otherwise timeout
        if (pin_state != 1) {
            // ubit.serial.send("Timeout waiting for Slave\r\n");
            return 6;
        }
        // ubit.serial.printf("Slave Latency: %i\r\n", end_time - start_time);
        // now, Slave should keep the Line HI for at least 400ms
        this->busy_wait_on_pin(1, 400, pin_state, start_time, end_time);
        // if the pin is LO, then Slave didn't drive it long enough
        // ubit.serial.printf("Slave Reply: %i\r\n", end_time - start_time);
        if (pin_state == 0) {
            // ubit.serial.send("Slave Reply too SHORT\r\n");
            return 8;
        }
        // otherwise, pin is still HI --allow Slave up to 200ms to bring it LO
        this->busy_wait_on_pin(1, 200, pin_state, start_time, end_time);
        // ubit.serial.printf("Slave Reply: %i\r\n", end_time - start_time);
        // if the pin is still HI, then Slave drove the pin for too long
        if (pin_state == 1) {
            // ubit.serial.send("Slave Reply too LONG\r\n");
            return 10;
        }
        // otherwise, all we need to do now is to wait for synchronisation
        return this->await_synchronisation();
    }

    int pair_as_slave() {
        // check the pin can read digital values
        int pin_state = ubit.io.pin[COMMS_PIN_NUMBER].getDigitalValue();
        if (pin_state == MICROBIT_NOT_SUPPORTED) {
            // ubit.serial.send("MICROBIT_NOT_SUPPORTED\r\n");
            return 1;
        } // oh no, it can't
        // otherwise, it can read digital values
        // now, wait up to 10000ms for Master to drive the line HI
        unsigned long start_time, end_time;
        this->busy_wait_on_pin(0, 10000, pin_state, start_time, end_time);
        // clarify the loop ended because pin is now indeed HI
        if (pin_state != 1) {
            // ubit.serial.send("Timeout waiting for Master\r\n");
            return 3;
        } // we timed out
        // otherwise, line is HI, wait for it to stay that way at least 400ms
        this->busy_wait_on_pin(1, 400, pin_state, start_time, end_time);
        // ubit.serial.printf("Master Probe: %i\r\n", end_time - start_time);
        // clarify the loop ended because timeout occurred (pin should be HI)
        if (pin_state != 1) {
            // ubit.serial.send("Master Probe too SHORT\r\n");
            return 5;
        } // pin went LO, that's not correct
        // wait up to 200 more ms for line to go LO
        this->busy_wait_on_pin(1, 200, pin_state, start_time, end_time);
        // ubit.serial.printf("Master Probe: %i\r\n", end_time - start_time);
        // if line is still HI, then it's overrun
        if (pin_state == 1) {
            // ubit.serial.send("Master Probe too LONG\r\n");
            return 7;
        }
        // otherwise, it's now our turn to bring the line HI
        // the most recent end_time is also the sync point, store that first
        this->synchronisation_timestamp = end_time;
        // drive line HI
        this->busy_wait_drive_pin(500, start_time, end_time);
        // now we just need to await synchronisation
        return this->await_synchronisation();
    }

    /**
     * @brief Drives the comms pin HI and performs a busy-wait loop for a given
     * duration, before driving the pin LO again.
     * @param duration length of time in ms to drive the pin HI for
     * @param start_time[out] system timestamp (ms) that the pin went HI
     * @param end_time[out] system timestamp (ms) that the pin went LO
     */
    void busy_wait_drive_pin(
        unsigned long duration,
        unsigned long& start_time,
        unsigned long& end_time
    ) {
        start_time = ubit.systemTime();
        end_time = start_time;
        ubit.io.pin[COMMS_PIN_NUMBER].setDigitalValue(1);
        // wait for specified duration
        do {
            /*
             * NOTE: this is a busy wait loop with no sleep()
             * this is okay as we're deliberately allowing this function to
             * block the entire system, it's documented as such and we also can
             * only sleep to a 6ms precision, but we want to time as accurately
             * as possible
             */
            end_time = ubit.systemTime();
        } while ((end_time - start_time) < duration);
        // bring line LO
        ubit.io.pin[COMMS_PIN_NUMBER].setDigitalValue(0);
    }

    /**
     * @brief Performs a busy-wait loop, continuously checking the comms pin
     * to make sure it is in a given state for a certain amount of time
     * @details method returns once either the pin is no longer in the desired
     * state, or if the given timeout has passed
     * @param keep_state the desired state (HI = 1 or LO = 0) that the pin is to
     * keep
     * @param timeout the maximum amount of time that the method will detect the
     * pin holding this state (and block) for
     * @param pin_state[out] the last detected pin state will be stored here
     * @param start_time[out] the system time that detection commenced (ms) will
     * be stored here
     * @param end_time[out] the system time that detection ended (ms) will be
     * stored here
     */
    void busy_wait_on_pin(
        int keep_state,
        unsigned long timeout,
        int& pin_state,
        unsigned long& start_time,
        unsigned long& end_time
    ) {
        // initialise time counters
        start_time = ubit.systemTime();
        end_time = start_time;
        // wait up to the specified timeout
        do {
            /*
             * NOTE: this is a busy wait loop with no sleep()
             * this is okay as we're deliberately allowing this function to
             * block the entire system, it's documented as such and we also can
             * only sleep to a 6ms precision, but we want to time as accurately
             * as possible
             */
            end_time = ubit.systemTime();
            pin_state = ubit.io.pin[COMMS_PIN_NUMBER].getDigitalValue();
        } while (pin_state == keep_state and (end_time - start_time) < timeout);
    }

    /**
     * @brief Blocks caller until 1250ms have elapsed since the stored
     * synchronisation timestamp
     * @returns true if the timestamp was in the future and the synchronisation
     * point has been reached
     * @returns false if the timestamp was not in the future (in this case,
     return is immediate)
     * @warn Don't call this if this->synchronisation_timestamp hasn't been set
     * first!
     */
    int await_synchronisation() {
        // calculate synchronisation time and check it's in the future
        unsigned long sync_time = this->synchronisation_timestamp + 2500;
        // ubit.serial.printf("Sync Stamp: %i\r\n", sync_time);
        unsigned long now_time = ubit.systemTime();
        if (sync_time < now_time) return 11; // can't do it, sync point in past
        // busy wait loop to wait until synchronisation time is reached
        do {
            now_time = ubit.systemTime();
        } while (now_time < sync_time);
        return 0; // at this point, now both Master and Slave are synchronised
    }

    unsigned long synchronisation_timestamp;
};


// a global MorseDecoder instance --ButtonTracker needs access to it so declare
MorseDecoder morse_decoder(MORSE_UNIT_MS);


/*
 * a very simple class that tracks the button state as it changes and the
 * duration between changes
 */
class ButtonTracker {
public:
    ButtonTracker()
        : current_state(false) // button always starts UP (not held down)
        , last_update(0) // initialise this to zero --no events yet
        , started_yet(false) // wait until our first UP/DOWN event
        {}

    /**
     * @brief Initialises this ButtonTracker, including attaching event-handler
     */
    void init() {
        // set up event handler(s)
        ubit.messageBus.listen(
            MICROBIT_ID_BUTTON_A,
            MICROBIT_EVT_ANY, // bind to all events so we get both UP and DOWN
            this,
            &ButtonTracker::on_button_a_change
        );
    }

    /**
     * @brief Tears down this ButtonTracker, including removing event-handler
     */
    void stop() {
        // set up event handler(s)
        ubit.messageBus.ignore(
            MICROBIT_ID_BUTTON_A,
            MICROBIT_EVT_ANY,
            this,
            &ButtonTracker::on_button_a_change
        );
    }

    /**
     * @returns true if input has timed out --this is useful for detecting word
     * spaces
     */
    bool has_timed_out() {
        return (ubit.systemTime() - this->last_update) >=
                morse_decoder.suggested_timeout_duration();
    }

    // instance method event handler for UP/DOWN events on button A
    // TODO: Shorten this event-handler! It seems to be doing far too much!
    void on_button_a_change(MicroBitEvent e) {
        // if we haven't yet started, we can only start on a DOWN event
        if (not this->started_yet) {
            if (e.value == MICROBIT_BUTTON_EVT_DOWN) {
                this->started_yet = true; // now we have started
                this->last_update = ubit.systemTime(); // store time of event
                this->current_state = true; // store event state
                // that's all --as this is the first event, we don't send it
            }
        } else { // once we've started, we evaluate events here from then-on
            // check event type --we only want UP or DOWN events
            switch (e.value) {
            case MICROBIT_BUTTON_EVT_UP:
                // only accept this event if previous state was DOWN
                if (this->current_state == true) {
                    unsigned long previous_update = this->last_update;
                    this->last_update = ubit.systemTime();
                    this->current_state = false;
                    // send the duration of the previous pulse to the decoder
                    morse_decoder.transition(true, this->last_update - previous_update);
                    // invert middle pixel
                    ubit.display.image.setPixelValue(2, 2, 255 - ubit.display.image.getPixelValue(2, 2));
                }
                break;
            case MICROBIT_BUTTON_EVT_DOWN:
                // only accept this event if previous state was UP
                if (this->current_state == false) {
                    unsigned long previous_update = this->last_update;
                    this->last_update = ubit.systemTime();
                    this->current_state = true;
                    // send the duration of the previous pulse to the decoder
                    morse_decoder.transition(false, this->last_update - previous_update);
                    // invert middle pixel
                    ubit.display.image.setPixelValue(2, 2, 255 - ubit.display.image.getPixelValue(2, 2));
                }
                break;
            }
        }
    }

    // TODO: consider providing destructor that will de-register the handler
private:
    bool current_state; // whether current button state is DOWN or UP
    unsigned long last_update; // system timestamp of last UP/DOWN change (ms)
    bool started_yet; // this flag isn't set to true until our first UP DOWN event
};


/**
 * @returns true if a ≈ b, within ±margin% of b (where margin = 1.0 is 100%)
 */
template <typename T>
bool is_roundabout(T a, T b, double margin) {
    T max_bound = b * (1.0 + margin);
    T min_bound = b * (1.0 - margin);
    return min_bound <= a && a <= max_bound;
}


/**
 * @brief Monitors the Line (pin input) for pulses, decodes them into bits and
 * sends them into the ManchesterDecoder.
 */
class LineMonitor {
public:
    /**
     * @brief Default constructor
     * @param decoder ManchesterDecoder instance to send the input bits into
     */
    LineMonitor(ManchesterDecoder& decoder)
        : decoder(decoder)
        , reception_in_progress(false)
        {}

    /**
     * @brief Initialises this LineMonitor, including attaching event-handler
     */
    void init() {
        // set input pin to generate pule events speicifically
        ubit.io.pin[COMMS_PIN_NUMBER].eventOn(MICROBIT_PIN_EVENT_ON_PULSE);
        // set up event handler(s) for both HI and LO pulse events
        ubit.messageBus.listen(
            COMMS_PIN_EVENT_ID,
            MICROBIT_PIN_EVT_PULSE_HI,
            this,
            &LineMonitor::on_pulse
        );
        ubit.messageBus.listen(
            COMMS_PIN_EVENT_ID,
            MICROBIT_PIN_EVT_PULSE_LO,
            this,
            &LineMonitor::on_pulse
        );
    }

    /**
     * @brief Tears down this LineMonitor, including removing event-handler
     */
    void stop() {
        // block all input events on the comms pin
        ubit.io.pin[COMMS_PIN_NUMBER].eventOn(MICROBIT_PIN_EVENT_NONE);
        // tear down event handlers
        ubit.messageBus.ignore(
            COMMS_PIN_EVENT_ID,
            MICROBIT_PIN_EVT_PULSE_HI,
            this,
            &LineMonitor::on_pulse
        );
        ubit.messageBus.ignore(
            COMMS_PIN_EVENT_ID,
            MICROBIT_PIN_EVT_PULSE_LO,
            this,
            &LineMonitor::on_pulse
        );
    }

    /**
     * @brief Event-handler for input pin pulse HI or LO events
     */
    void on_pulse(MicroBitEvent e) {
        // create a new Pulse and store it in the buffer
        Pulse p = {
            e.value == MICROBIT_PIN_EVT_PULSE_LO ? false : true, // state
            e.timestamp // duration
        };
        // if it's a long silence, this is a good time to clear input buffer
        if (p.state == false and p.duration > TransceiverConfig::MAXIMUM_PULSE_DURATION) {
            this->decoder.clear(); // this helps mitigate bit-slip errors persisting
            // ubit.serial.send("Clear Buffer\r\n");
        }
        // otherwise only act on pulses which are shorter than a certain duration
        if (p.duration > TransceiverConfig::MAXIMUM_PULSE_DURATION) return;
        // ubit.serial.printf("on_pulse() -> { %i, %lu, }\r\n", p.state, p.duration);
        this->buffer.add(p);
    }

    /**
     * @brief Decodes all currently-buffered input pulses.
     * @details Call this regularly to keep processing input data and keep in
     * sync.
     * STOP bits
     */
    void flush() {
        bool has_output = false;
        for (size_t i = 0; i < this->buffer.length(); i++) {
            Pulse p = this->buffer[i];
            // attempt to decode the Pulse
            // handle all remaining HI/LO bits according to duration
            if (
                is_roundabout(
                    p.duration,
                    TransceiverConfig::SIGNAL_BIT_DURATION * 2, // double-length bits
                    TransceiverConfig::DURATION_MARGIN
                )
            ) {
                // send two inputs of the given state
                has_output = true;
                // ubit.serial.printf("%d%d", p.state, p.state);
                this->decoder.input(p.state);
                this->decoder.input(p.state);
            } else if (
                is_roundabout(
                    p.duration,
                    TransceiverConfig::SIGNAL_BIT_DURATION, // single-length bits
                    TransceiverConfig::DURATION_MARGIN
                )
            ) {
                // send just one input of the given state
                has_output = true;
                // ubit.serial.printf("%d", p.state);
                this->decoder.input(p.state);
            }
        }
        if (has_output) {
            // ubit.serial.send("<- MANCHEDECODE in\r\n");
        }
        // empty the buffer
        this->buffer.clear();
    }
private:
    ManchesterDecoder& decoder; /**< ManchesterDecoder to send bits into */

    bool reception_in_progress; /**< Whether a transmission is currently being received */

    CircularQueueBuffer<Pulse, 72u> buffer; // buffer with 72 spaces available
};


/**
 * @brief The opposite of LineMonitor, this class takes bit stream input and
 * switches the Comms pin state accordingly to transmit the stream of bits out
 * on the Line.
 */
class LineDriver {
public:
    /**
     * @brief Default constructor
     */
    LineDriver()
        {}

    /**
     * @brief Send an input bit into the LineDriver
     * @param state the input bit to send
     * @details the input is not sent immediately but queued for sending. Call
     * flush() to send all bits when buffer is ready.
     * @returns true if bit could be added to buffer
     * @returns false if buffer is full
     */
    bool input(bool state) {
        return this->buffer.add(state);
    }

    /**
     * @brief Flushes the output buffer out to the Line
     * @note This method will block the caller!
     * @returns true if buffer flushed
     * @returns false if buffer isn't big enough yet (can't flush)
     * @todo Consider using more accurate timing than just sleep()
     */
    bool flush() {
        if (this->buffer.length() < 16u) return false;
        // START BIT --should be inverted version of first bit
        ubit.io.pin[COMMS_PIN_NUMBER].setDigitalValue(not this->buffer[0]);
        ubit.sleep(TransceiverConfig::STOP_START_BIT_DURATION_MS);
        // BEGIN DATA BITS
        for (size_t i = 0; i < 16u; i++) {
            ubit.io.pin[COMMS_PIN_NUMBER].setDigitalValue(this->buffer[i]);
            ubit.sleep(TransceiverConfig::SIGNAL_BIT_DURATION_MS);
        }
        // END DATA BITS
        // STOP BIT --should be inverted version of last bit
        ubit.io.pin[COMMS_PIN_NUMBER].setDigitalValue(not this->buffer[15u]);
        ubit.sleep(TransceiverConfig::STOP_START_BIT_DURATION_MS);
        // drive low then switch to input mode
        ubit.io.pin[COMMS_PIN_NUMBER].setDigitalValue(0);
        ubit.io.pin[COMMS_PIN_NUMBER].getDigitalValue();
        // remove transmitted bits
        this->buffer.shrink_from_front(16u);
        return true; // no problem
    }
private:
    CircularQueueBuffer<bool, 64u> buffer; // buffer with enough space for 4 bytes
};


// this will track the button presses
ButtonTracker button_tracker;
// this is used to encode bytes we want to send to go out on the line
ManchesterEncoder manchester_encoder;
// this switches the line state
LineDriver line_driver;
// this is used to decode bytes from the line
ManchesterDecoder manchester_decoder;
// this monitors the line for changes
LineMonitor line_monitor(manchester_decoder);


// flag used by the morse_checker() fiber function to check if it should run
bool morse_checker_enabled = false;


// simple fiber used for checking Morse decode state (for now), will send in future
void morse_checker() {
    char current = '#'; // currently-seen char from Morse input
    while (true) {
        ubit.sleep(100); // yield the CPU for a bit
        // guard clause preventing execution of the remainder unless enabled
        if (not morse_checker_enabled) continue;
        // if input has timed out, send a word space transition
        if (button_tracker.has_timed_out()) {
            morse_decoder.transition(false, morse_decoder.suggested_timeout_duration());
        }
        char previous = current; // track what last-seen char was
        if (morse_decoder.remove_next(current)) { // if we could read another Morse code
            // current has now updated, so display this new char
            ubit.display.printCharAsync(current, 500);
            // whenever another character is removed, send it down the line
            /*
             * NOTE: for now, just sending characters as-is but in the future
             * they will be encoded rather than sent raw
             */
            // ubit.serial.printf("Morse Input: %c\r\n", current);
            manchester_encoder.input(current); // NOTE: Might need cast
            ubit.display.clear();
        } else {
            // otherwise, peek into what the code appears to be thus far
            char peek = morse_decoder.peek();
            // if this peek is different to what it previously was, display
            if (peek != current) {
                current = peek;
                ubit.display.printChar(current);
            }
        }
    }
}


// simple fiber that simply passes data out of encoder and into the line driver
void flush_output() {
    while (true) {
        ubit.sleep(10); // yield the CPU for a bit
        int next_state = 0;
        bool printed = false;
        while (manchester_encoder.get_next_state(next_state)) {
            printed = true;
            // ubit.serial.printf("%i", next_state);
            line_driver.input(next_state);
        }
        if (printed) {
            // ubit.serial.send(" <- MANCHECODE out\r\n");
        }
    }
}


// simple fiber that simply flushes data out of line monitor and into decoder
void flush_input() {
    while (true) {
        ubit.sleep(10); // yield the CPU for a bit
        line_monitor.flush();
    }
}


int main() {
    // Initialise the micro:bit runtime.
    ubit.init();

    // greyscale mode please!
    ubit.display.setDisplayMode(DISPLAY_MODE_GREYSCALE);

    // these lightweight fibers ferry data to and from decoders and line stuff

    // this is the main program loop, an infinite while!
    while (true) {
        // first of all, work out what Role we are to play
        Role my_role = get_my_role(); // this call blocks and asks the user

        Pairer pairer;

        // try to pair and if unsuccessful, go back to top of loop to restart
        int error_code = pairer.attempt_pair(my_role);
        if (error_code != 0) {
            // scroll "PF" to indicate "PAIR FAILURE"
            ubit.display.scroll("PF");
            ubit.display.scroll(error_code);
            continue; // skip back to start of loop body to re-try pair
        }
        // otherwise, if execution continues here then we're paired

        // for now, we only act on button presses if we're the Master
        if (my_role == Role::MASTER) {
            // scroll "PM" to mean "PAIRED MASTER"
            // ubit.serial.send(" -> Role::MASTER\r\n");
            ubit.display.scroll("PM");
        } else if (my_role == Role::SLAVE) {
            // scroll "PS" to mean "PAIRED SLAVE"
            // ubit.serial.send(" -> Role::SLAVE\r\n");
            ubit.display.scroll("PS");
        }

        /* BEGIN COMMS */

        if (my_role == Role::MASTER) { // Master's comms loop
            // initialise the ButtonTracker
            button_tracker.init();
            // create one fiber --this infinitely checks the decoder and outputs changes
            create_fiber(morse_checker);
            // enable the morse checker!
            morse_checker_enabled = true;
            create_fiber(flush_output);
            // line_driver.init();
            while (true) { // TODO: consider what exit conditions can be used
                // call barrier to synchronise both master and slave
                pairer.barrier();
                // flush the line driver
                line_driver.flush();
            }
        } else { // Slave's comms loop
            create_fiber(flush_input);
            line_monitor.init();
            while (true) { // TODO: consider what exit conditions can be used
                // call barrier to synchronise both master and slave
                pairer.barrier();
                // TODO: display the output?
                uint8_t next_byte = 0x00u;
                if (manchester_decoder.get_next_byte(next_byte)) {
                    // ubit.serial.printf("MANCHEDECODE -> %c\r\n", next_byte);
                    ubit.display.printChar(next_byte, 300);
                    ubit.display.clear();
                }
            }
        }

        /* END COMMS */

        /*
         * NOTE: the close-down code here is currently unreachable but will be
         * in the future
         */

        // transceiver.close();

        // if we're the Master, we need to stop the button tracker
        if (my_role == Role::MASTER) {
            // button_tracker.stop();
            // disable the morse checker
            morse_checker_enabled = false;
        }
    }

    // TODO: potentially remove this call, if clarified that it is not required.
    // If main exits, there may still be other fibers running or registered event handlers etc.
    // Simply release this fiber, which will mean we enter the scheduler. Worse case, we then
    // sit in the idle task forever, in a power efficient sleep.
    release_fiber();
}
